// Math.h - Basic math algorithms.
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.02 - 09/19/12 - Add GetSign
// - v1.01 - 09/15/12 - Add IsPow2
// - v1.00 - 06/07/12 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - 

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#pragma once

#include <math.h>

namespace glory {

const float FUZZY_PLACES_F[10] = { 1.0f, 0.1f, 0.01f, 0.001f, 0.0001f, 0.00001f, 0.000001f, 0.0000001f, 0.00000001f, 0.000000001f };
const double FUZZY_PLACES_D[10] = { 1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001, 0.000001, 0.0000001, 0.00000001, 0.000000001 };

// For float operations
template <int P> inline bool FuzzyEq(float a, float b) { return fabs(a - b) <= FUZZY_PLACES_F[P]; }
template <int P> inline bool FuzzyNe(float a, float b) { return !FuzzyEq<P>(a, b); }
template <int P> inline bool FuzzyGt(float a, float b) { return a > (b + FUZZY_PLACES_F[P]); }
template <int P> inline bool FuzzyGe(float a, float b) { return a >= (b - FUZZY_PLACES_F[P]); }
template <int P> inline bool FuzzyLt(float a, float b) { return a < (b - FUZZY_PLACES_F[P]); }
template <int P> inline bool FuzzyLe(float a, float b) { return a <= (b + FUZZY_PLACES_F[P]); }
inline bool FuzzyEq(float a, float b) { return FuzzyEq<5>(a, b); }
inline bool FuzzyNe(float a, float b) { return FuzzyNe<5>(a, b); }
inline bool FuzzyGt(float a, float b) { return FuzzyGt<5>(a, b); }
inline bool FuzzyGe(float a, float b) { return FuzzyGe<5>(a, b); }
inline bool FuzzyLt(float a, float b) { return FuzzyLt<5>(a, b); }
inline bool FuzzyLe(float a, float b) { return FuzzyLe<5>(a, b); }
template <int P> inline bool FuzzyEq(double a, double b) { return fabs(a - b) <= FUZZY_PLACES_D[P]; }
template <int P> inline bool FuzzyNe(double a, double b) { return !FuzzyEq<P>(a, b); }
template <int P> inline bool FuzzyGt(double a, double b) { return a > (b + FUZZY_PLACES_D[P]); }
template <int P> inline bool FuzzyGe(double a, double b) { return a >= (b - FUZZY_PLACES_D[P]); }
template <int P> inline bool FuzzyLt(double a, double b) { return a < (b - FUZZY_PLACES_D[P]); }
template <int P> inline bool FuzzyLe(double a, double b) { return a <= (b + FUZZY_PLACES_D[P]); }
inline bool FuzzyEq(double a, double b) { return FuzzyEq<5>(a, b); }
inline bool FuzzyNe(double a, double b) { return FuzzyNe<5>(a, b); }
inline bool FuzzyGt(double a, double b) { return FuzzyGt<5>(a, b); }
inline bool FuzzyGe(double a, double b) { return FuzzyGe<5>(a, b); }
inline bool FuzzyLt(double a, double b) { return FuzzyLt<5>(a, b); }
inline bool FuzzyLe(double a, double b) { return FuzzyLe<5>(a, b); }

// 2 * pi radians in a circle
inline float RadToRad(float rad) { while (rad < 0) rad += 3.1415927f * 2; while (rad >= 3.1415927f * 2) rad -= 3.1415927f * 2; return rad; }
inline float DegToDeg(float deg) { while (deg < 0) deg += 360.0f; while (deg >= 360.0f) deg -= 360.0f; return deg; }
inline float CustomToCustom(float custom, float customMax) { while (custom < 0) custom += customMax; while (custom >= customMax) custom -= customMax; return custom; }
inline float DegToRad(float deg) { return RadToRad(deg / 180.0f * 3.1415927f); }
inline float RadToDeg(float rad) { return DegToDeg(rad / 3.1415927f * 180.0f); }
inline float CustomToRad(float custom, float customMax) { return RadToRad((custom / (customMax / 2.0f)) * 3.1415927f); }
inline float RadToCustom(float rad, float customMax) { return CustomToCustom((rad / 3.1415927f) * (customMax / 2.0f), customMax); }
inline double RadToRad(double rad) { while (rad < 0) rad += 3.1415927 * 2; while (rad >= 3.1415927 * 2) rad -= 3.1415927 * 2; return rad; }
inline double DegToDeg(double deg) { while (deg < 0) deg += 360.0; while (deg >= 360.0) deg -= 360.0; return deg; }
inline double CustomToCustom(double custom, double customMax) { while (custom < 0) custom += customMax; while (custom >= customMax) custom -= customMax; return custom; }
inline double DegToRad(double deg) { return RadToRad(deg / 180.0 * 3.1415927); }
inline double RadToDeg(double rad) { return DegToDeg(rad / 3.1415927 * 180.0); }
inline double CustomToRad(double custom, double customMax) { return RadToRad((custom / (customMax / 2.0)) * 3.1415927); }
inline double RadToCustom(double rad, double customMax) { return CustomToCustom((rad / 3.1415927) * (customMax / 2.0), customMax); }

inline int GetGcd(int a, int b) {
	int rem = a % b;
	while (rem) {
		a = b;
		b = rem;
		rem = a % b;
	}
	return b;
}

inline int GetLcm(int a, int b) {
	return abs(a / GetGcd(a, b) * b);
}

template <class T> inline T GetClamped(T val, T minVal, T maxVal) { 
	return val < minVal ? minVal : (val > maxVal ? maxVal : val); 
}

template <class T> inline T GetSign(T val) { if (val < 0) return static_cast<T>(-1); return static_cast<T>(1); }

template <class T> inline bool IsSignEqual(T a, T b) { return (a >= 0) == (b >= 0); }

// Specifically for non-float types
template <class T> inline bool IsPow2(T a) { return !(a & (a - 1)); }
inline int GetNextPow2(int a) {
	if (a-- < 0)
		return 0;
	a |= a >> 1;
	a |= a >> 2;
	a |= a >> 4;
	a |= a >> 8;
	a |= a >> 16;
	return a + 1;
}
inline unsigned int GetNextPow2(unsigned int a) {
	a--;
	a |= a >> 1;
	a |= a >> 2;
	a |= a >> 4;
	a |= a >> 8;
	a |= a >> 16;
	return a + 1;
}
inline long long GetNextPow2(long long a) {
	if (a-- < 0)
		return 0;
	a |= a >> 1;
	a |= a >> 2;
	a |= a >> 4;
	a |= a >> 8;
	a |= a >> 16;
	a |= a >> 32;
	return a + 1;
}
inline unsigned long long GetNextPow2(unsigned long long a) {
	a--;
	a |= a >> 1;
	a |= a >> 2;
	a |= a >> 4;
	a |= a >> 8;
	a |= a >> 16;
	a |= a >> 32;
	return a + 1;
}

// Assigns the specified value to min if it is less, and max if it is greater
template <class T> inline void AssignMinMax(T &currMin, T &currMax, const T &val) {
	if (val < currMin) 
		currMin = val;
	if (val > currMax) 
		currMax = val;
}

inline float Round(float r) { return floor(r + 0.5f); }
inline double Round(double r) { return floor(r + 0.5); }
inline float RoundToNearest(float r, float n) { float rem = fmod(r, n); if (rem / n < .5f) return r - rem; return r - rem + n; }
inline double RoundToNearest(double r, double n) { double rem = fmod(r, n); if (rem / n < .5) return r - rem; return r - rem + n; }

}  // namespace glory
