// Random.h - Fast random number generator.
// It is a type of LGC of the CMWC 4096 breed.
// Based on discussion from http://www.gamedev.net/topic/512260-number-generator-cmwc-4096-and-implementation/
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.01 - 09/28/12 - Fixed typo that creeped in default c'tor
//                    - Ensure correct built in primitive types
// - v1.00 - 05/31/12 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - 

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#pragma once

namespace glory {

class Random {
public:
	Random() : mIndex(4095), mC(7919) { Seed(0); }
	explicit Random(int seedVal) : mIndex(4095), mC(7919) { Seed(seedVal); }
	~Random() { }

	static inline Random &GetGlobalInstance() {
		static Random instance;
		return instance;
	};

	inline void Seed(int seedVal) {
		// This is actually another random number generator code...
		if (seedVal == 0)
			seedVal = mC;
		int val = seedVal;
		for (int i = 0; i < 4096; i++) {
			val *= 16807;
			mQ[i] = val;
		}
	}

	inline unsigned int GetRandom32u() { 
		mIndex = (mIndex + 1) & 4095; // Advance through the state array, circularly
		unsigned long long t = 18782ull * mQ[mIndex] + mC; // Notice that the result is computed as a 64-bit integer.
		mC = (t >> 32); // except we keep changing the constant we add.
		t = (t & 0xffffffffull) + mC; // add the lower 32 bits with the higher 32 bits to avoid poor quality bits at the extremes.
		if (t > 0xfffffffeull) // About 50% of the time (?? I don't know why this helps).
		{
			t -= 0xffffffffull;
			++mC;
		}
		return (mQ[mIndex] = static_cast<unsigned int>(0xfffffffeull - t)); // I don't know why this is better than simply `(Q[i] = x)'.
	}

	inline unsigned int GetRandom32u(unsigned int maxVal) {
		return GetRandom32u() % (maxVal + 1);
	}

	inline int GetRandom32i() { return static_cast<int>(GetRandom32u()); }

	// Returns a number from 0 (inclusive) to 1 (exclusive)
	inline float GetRandomFloat() {
		union {
			unsigned long iVal;
			float fVal;
		} Conversion;
		// seee eeee efff ffff ffff ffff ffff ffff
		// float value = s * 2^(e-127) * m, where m = 1.f, and thus 1 <= m < 2
		Conversion.iVal = 0x3f800000 | (GetRandom32u() & 0x007fffff);
		return Conversion.fVal - 1.0f;
	}

	inline bool GetRandomBool() { return !!(GetRandom32u() & 0x100); }

	inline int GetRandomSign() { return GetRandomBool() ? 1 : -1; }

private:
	unsigned long mIndex;
	unsigned long mQ[4096]; // Q holds the state
	unsigned long mC;  // Constant
};

}  // namespace glory
